@@ -1,4 +1,4 @@
-(defproject puppetlabs/trapperkeeper-metrics "1.2.1"
+(defproject puppetlabs/trapperkeeper-metrics "1.2.2"
   :description "Trapperkeeper Metrics Service"
   :url "http://github.com/puppetlabs/trapperkeeper-metrics"
 
@@ -35,9 +35,9 @@
   :source-paths  ["src/clj"]
   :java-source-paths  ["src/java"]
 
-  :deploy-repositories [["releases" {:url "https://artifactory.delivery.puppetlabs.net/artifactory/clojure-releases__local/"
-                                     :username :env/nexus_jenkins_username
-                                     :password :env/nexus_jenkins_password
+  :deploy-repositories [["releases" {:url "https://clojars.org/repo"
+                                     :username :env/clojars_jenkins_username
+                                     :password :env/clojars_jenkins_password
                                      :sign-releases false}]]
 
   :classifiers  [["test" :testutils]]
