@@ -58,6 +58,7 @@
             [puppetlabs.puppetdb.catalogs :as cat]
             [puppetlabs.puppetdb.reports :as report]
             [puppetlabs.puppetdb.facts :as fact]
+            [puppetlabs.puppetdb.nodes :as nodes]
             [puppetlabs.kitchensink.core :as kitchensink]
             [puppetlabs.puppetdb.cheshire :as json]
             [puppetlabs.puppetdb.jdbc :as jdbc]
@@ -106,6 +107,7 @@
 ;;         :queue-time <histogram>
 ;;         :retry-counts <histogram>
 ;;         :size <histogram>
+;;         :ignored <counter>
 ;;        }
 ;;      "command name"
 ;;        {<version number>
@@ -116,6 +118,7 @@
 ;;            :discarded <meter>
 ;;            :processing-time <timer>
 ;;            :retry-counts <histogram>
+;;            :ignored <counter>
 ;;           }
 ;;        }
 ;;     }
@@ -140,6 +143,7 @@
 ;; * `:invalidated`: commands marked as delete?, caused by a newer command
 ;;                   was enqueued that will overwrite an existing one in the queue
 ;; * `:depth`: number of commands currently enqueued
+;; * `:ignored`: number of obsolete commands that have been ignored.
 ;;
 
 (def mq-metrics-registry (get-in metrics/metrics-registries [:mq :registry]))
@@ -159,7 +163,8 @@
      :fatal (meter mq-metrics-registry (to-metric-name-fn :fatal))
      :retried (meter mq-metrics-registry (to-metric-name-fn :retried))
      :awaiting-retry (counter mq-metrics-registry (to-metric-name-fn :awaiting-retry))
-     :discarded (meter mq-metrics-registry (to-metric-name-fn :discarded))}))
+     :discarded (meter mq-metrics-registry (to-metric-name-fn :discarded))
+     :ignored (counter mq-metrics-registry (to-metric-name-fn :ignored))}))
 
 (def metrics
   (atom {:command-parse-time (timer mq-metrics-registry
@@ -345,7 +350,6 @@
 (defn deactivate-node-wire-v1->wire-3 [deactive-node]
   (-> deactive-node
       (json/parse-string true)
-      upon-error-throw-fatality
       deactivate-node-wire-v2->wire-3))
 
 (defn deactivate-node*
@@ -359,12 +363,15 @@
     (log-command-processed-messsage id received start-time :deactivate-node certname)))
 
 (defn deactivate-node [{:keys [payload version] :as command} start-time db]
-  (-> command
-      (assoc :payload (case version
-                        1 (deactivate-node-wire-v1->wire-3 payload)
-                        2 (deactivate-node-wire-v2->wire-3 payload)
-                        payload))
-      (deactivate-node* start-time db)))
+  (let [validated-payload (upon-error-throw-fatality
+                           (s/validate nodes/deactivate-node-wireformat-schema
+                                       (case version
+                                         1 (deactivate-node-wire-v1->wire-3 payload)
+                                         2 (deactivate-node-wire-v2->wire-3 payload)
+                                         payload)))]
+    (-> command
+        (assoc :payload validated-payload)
+        (deactivate-node* start-time db))))
 
 ;; Report submission
 
