@@ -1,4 +1,6 @@
-(ns puppetlabs.puppetdb.command.constants)
+(ns puppetlabs.puppetdb.command.constants
+  (:require [clojure.set :as set]
+            [clojure.string :as str]))
 
 (def command-names
   {:replace-catalog "replace catalog"
@@ -6,6 +8,8 @@
    :deactivate-node "deactivate node"
    :store-report    "store report"})
 
+(def command-keys (set/map-invert command-names))
+
 (defn- version-range [min-version max-version]
   (set (range min-version (inc max-version))))
 
@@ -18,8 +22,13 @@
 (def latest-catalog-version (apply max (get supported-command-versions "replace catalog")))
 (def latest-report-version (apply max (get supported-command-versions "store report")))
 (def latest-facts-version (apply max (get supported-command-versions "replace facts")))
+(def latest-deactivate-node-version (apply max (get supported-command-versions "deactivate node")))
 
 (def latest-command-versions
   {:replace_catalog latest-catalog-version
    :store_report latest-report-version
    :replace_facts latest-facts-version})
+
+(defn normalize-command-name [command]
+  "Normalize command name from an incoming request's query param"
+  (str/replace command "_" " "))
