@@ -60,9 +60,11 @@
      :classname (pls/defaulted-maybe String "org.postgresql.Driver")
      :subprotocol (pls/defaulted-maybe String "postgresql")
      :subname (s/maybe String)
-     :username String
      :user String
+     :username String
      :password String
+     :migrator-username String
+     :migrator-password String
      :syntax_pgs String
      :read-only? (pls/defaulted-maybe String "false")
      :partition-conn-min (pls/defaulted-maybe s/Int 1)
@@ -73,7 +75,8 @@
      :statements-cache-size (pls/defaulted-maybe s/Int 0)
      :connection-timeout (pls/defaulted-maybe s/Int 3000)
      :facts-blacklist pls/Blacklist
-     :facts-blacklist-type (pls/defaulted-maybe (s/enum "literal" "regex") "literal")}))
+     :facts-blacklist-type (pls/defaulted-maybe (s/enum "literal" "regex") "literal")
+     :schema-check-interval (pls/defaulted-maybe s/Int (* 30 1000))}))
 
 (def write-database-config-in
   "Includes the common database config params, also the write-db specific ones"
@@ -83,7 +86,8 @@
             :report-ttl (pls/defaulted-maybe String "14d")
             :node-purge-ttl (pls/defaulted-maybe String "14d")
             :node-purge-gc-batch-limit (pls/defaulted-maybe s/Int 25)
-            :node-ttl (pls/defaulted-maybe String "7d")})))
+            :node-ttl (pls/defaulted-maybe String "7d")
+            :migrate (pls/defaulted-maybe String "true")})))
 
 (def database-config-out
   "Schema for parsed/processed database config"
@@ -103,12 +107,15 @@
    :connection-timeout s/Int
    :maximum-pool-size s/Int
    (s/optional-key :conn-lifetime) (s/maybe Minutes)
-   (s/optional-key :username) String
    (s/optional-key :user) String
+   (s/optional-key :username) String
    (s/optional-key :password) String
+   (s/optional-key :migrator-username) String
+   (s/optional-key :migrator-password) String
    (s/optional-key :syntax_pgs) String
    (s/optional-key :facts-blacklist) clojure.lang.PersistentVector
-   :facts-blacklist-type String})
+   :facts-blacklist-type String
+   :schema-check-interval s/Int})
 
 (def write-database-config-out
   "Schema for parsed/processed database config that includes write database params"
@@ -117,7 +124,8 @@
           :report-ttl Period
           :node-purge-ttl Period
           :node-purge-gc-batch-limit (s/constrained s/Int (complement neg?))
-          :node-ttl Period}))
+          :node-ttl Period
+          :migrate Boolean}))
 
 (defn half-the-cores*
   "Function for computing half the cores of the system, useful
@@ -263,6 +271,36 @@
          (s/validate database-config-out)
          (assoc config :read-database))))
 
+(defn prefer-db-user-on-username-mismatch
+  [{{:keys [user username]} :database :as config}]
+  ;; match puppetdb.jdbc/make-connection-pool
+  (when (and user username (not= user username))
+    (log/warn
+     (trs "Configured database user {0} and username {1} don't match"
+          (pr-str user) (pr-str username)))
+    (log/warn
+     (trs "Preferring configured user {0}" (pr-str user))))
+  (let [user (or user username)]
+    (-> config
+        (assoc-in [:database :user] user)
+        (assoc-in [:database :username] user))))
+
+(defn configure-db
+  [config]
+  (letfn [(ensure-migrator-info [config]
+            (let [username (get-in config [:database :username])]
+              (-> config
+                  (update-in [:database :migrator-username] #(or % username))
+                  (update-in [:database :migrator-password]
+                             #(or % (get-in config [:database :password]))))))]
+    (-> config
+        (configure-section :database
+                           write-database-config-in
+                           write-database-config-out)
+        prefer-db-user-on-username-mismatch
+        ensure-migrator-info
+        configure-read-db)))
+
 (defn configure-puppetdb
   "Validates the [puppetdb] section of the config"
   [{:keys [puppetdb] :as config :or {puppetdb {}}}]
@@ -295,8 +333,7 @@
    to the internal Clojure format that PuppetDB expects"
   [config]
   (-> config
-      (configure-section :database write-database-config-in write-database-config-out)
-      configure-read-db
+      configure-db
       configure-command-processing
       configure-puppetdb))
 
