@@ -5,7 +5,9 @@
             [puppetlabs.puppetdb.query.paging :as paging]
             [puppetlabs.puppetdb.query-eng :refer [produce-streaming-body
                                                    stream-query-result]]
-            [puppetlabs.puppetdb.middleware :refer [validate-query-params
+            [puppetlabs.puppetdb.middleware :refer [merge-param-specs
+                                                    params-schema
+                                                    validate-query-params
                                                     parent-check
                                                     handler-schema]]
             [puppetlabs.comidi :as cmdi]
@@ -15,19 +17,15 @@
             [puppetlabs.kitchensink.core :as kitchensink]
             [puppetlabs.puppetdb.scf.storage-utils :as sutils]
             [puppetlabs.puppetdb.utils :refer [assoc-when]]
-            [clojure.walk :refer [keywordize-keys]]))
-
-(def params-schema {(s/optional-key :optional) [s/Str]
-                    (s/optional-key :required) [s/Str]})
+            [clojure.walk :refer [keywordize-keys]]
+            [puppetlabs.i18n.core :refer [trs]]))
 
 ;; General route/handler construction functions
 
 (pls/defn-validated extract-query :- bidi-schema/RoutePair
-  ([routes :- bidi-schema/RoutePair]
-   (extract-query {:optional paging/query-params} routes))
-  ([param-spec :- params-schema
-    routes :- bidi-schema/RoutePair]
-   (cmdi/wrap-routes routes #(http-q/extract-query % param-spec))))
+  [param-spec :- params-schema
+   routes :- bidi-schema/RoutePair]
+  (cmdi/wrap-routes routes #(http-q/extract-query % param-spec)))
 
 (pls/defn-validated append-handler :- bidi-schema/RoutePair
   [route :- bidi-schema/RoutePair
@@ -52,14 +50,6 @@
    route-param-key :- s/Keyword]
   (cmdi/wrap-routes route #(parent-check % version entity route-param-key)))
 
-(pls/defn-validated experimental-root-routes :- bidi-schema/RoutePair
-  [version :- s/Keyword]
-  (cmdi/ANY "" []
-            (-> (http-q/query-handler version)
-                (http-q/extract-query-pql {:optional (conj paging/query-params "ast_only")
-                                           :required ["query"]})
-                (http/experimental-warning "The root endpoint is experimental"))))
-
 (defn report-data-responder
   "Respond with either metrics or logs for a given report hash.
    `entity` should be either :metrics or :logs."
@@ -140,16 +130,40 @@
 
 ;; Routes
 
+;; query parameter sets
+(def global-params {:optional []})
+(def paging-params {:optional paging/query-params})
+(def pretty-params {:optional ["pretty"]})
+(def typical-params (merge-param-specs global-params
+                                       paging-params
+                                       pretty-params
+                                       {:optional ["include_package_inventory"]}))
+
+(pls/defn-validated root-routes :- bidi-schema/RoutePair
+  [version :- s/Keyword]
+  (let [no-certname-entities #{"fact_paths" "environments" "packages"}
+        get-entity #(some-> % :puppetdb-query :query second)]
+    (cmdi/ANY "" []
+              (-> (comp (http-q/query-handler version)
+                        (fn [req]
+                          (cond
+                            (some-> req :puppetdb-query :ast_only http-q/coerce-to-boolean) req
+                            (no-certname-entities (get-entity req)) req
+                            :else (http-q/restrict-query-to-active-nodes req))))
+                  (http-q/extract-query-pql
+                   (merge-param-specs typical-params
+                                      {:optional ["ast_only"]
+                                       :required ["query"]}))))))
+
 (pls/defn-validated events-routes :- bidi-schema/RoutePair
   "Ring app for querying events"
   [version :- s/Keyword]
   (extract-query
-   {:optional (concat
-               ["query"
-                "distinct_resources"
-                "distinct_start_time"
-                "distinct_end_time"]
-               paging/query-params)}
+   (merge-param-specs typical-params
+                      {:optional ["query"
+                                  "distinct_resources"
+                                  "distinct_start_time"
+                                  "distinct_end_time"]})
    (cmdi/routes
     (cmdi/ANY "" []
               (create-query-handler version "events")))))
@@ -158,6 +172,7 @@
   [version :- s/Keyword]
   (cmdi/routes
    (extract-query
+    typical-params
     (cmdi/ANY "" []
               (create-query-handler version "reports")))
 
@@ -169,16 +184,19 @@
    (cmdi/ANY ["/" :hash "/metrics"] []
              (-> (report-data-responder version "report_metrics")
                  (parent-check version :report :hash)
-                 (validate-query-params {:optional ["pretty"]})))
+                 (validate-query-params (merge-param-specs global-params
+                                                           pretty-params))))
 
    (cmdi/ANY ["/" :hash "/logs"] []
              (-> (report-data-responder version "report_logs")
                  (parent-check version :report :hash)
-                 (validate-query-params {:optional ["pretty"]})))))
+                 (validate-query-params (merge-param-specs global-params
+                                                           pretty-params))))))
 
 (pls/defn-validated resources-routes :- bidi-schema/RoutePair
   [version :- s/Keyword]
   (extract-query
+   typical-params
    (cmdi/routes
     (cmdi/ANY "" []
               (create-query-handler version "resources"  http-q/restrict-query-to-active-nodes))
@@ -197,12 +215,14 @@
 (pls/defn-validated edge-routes :- bidi-schema/RoutePair
   [version :- s/Keyword]
   (extract-query
+   typical-params
    (cmdi/ANY "" []
              (create-query-handler version "edges" http-q/restrict-query-to-active-nodes))))
 
 (pls/defn-validated catalog-routes :- bidi-schema/RoutePair
   [version :- s/Keyword]
   (extract-query
+   typical-params
    (cmdi/routes
 
     (cmdi/ANY "" []
@@ -224,6 +244,7 @@
 (pls/defn-validated facts-routes :- bidi-schema/RoutePair
   [version :- s/Keyword]
   (extract-query
+   typical-params
    (cmdi/routes
     (cmdi/ANY "" []
               (create-query-handler version "facts" http-q/restrict-query-to-active-nodes))
@@ -244,14 +265,36 @@
 (pls/defn-validated inventory-routes :- bidi-schema/RoutePair
   [version :- s/Keyword]
   (extract-query
+    typical-params
     (cmdi/routes
       (cmdi/ANY "" []
                 (create-query-handler version "inventory"
                                       http-q/restrict-query-to-active-nodes)))))
 
+(pls/defn-validated packages-routes :- bidi-schema/RoutePair
+  [version :- s/Keyword]
+  (extract-query
+   typical-params
+   (cmdi/routes
+    (cmdi/ANY "" []
+              (create-query-handler version "packages")))))
+
+(pls/defn-validated package-inventory-routes :- bidi-schema/RoutePair
+  [version :- s/Keyword]
+  (extract-query
+    typical-params
+    (cmdi/routes
+      (cmdi/ANY "" []
+                (create-query-handler version "package_inventory"
+                                      http-q/restrict-query-to-active-nodes))
+      (cmdi/context ["/" :node]
+                    (cmdi/ANY "" [] (create-query-handler version "package_inventory"
+                                                          http-q/restrict-query-to-node))))))
+
 (pls/defn-validated factset-routes :- bidi-schema/RoutePair
   [version :- s/Keyword]
   (extract-query
+   typical-params
    (cmdi/routes
     (cmdi/ANY "" []
               (create-query-handler version "factsets" http-q/restrict-query-to-active-nodes))
@@ -267,6 +310,7 @@
 (pls/defn-validated fact-names-routes :- bidi-schema/RoutePair
   [version :- s/Keyword]
   (extract-query
+   typical-params
    (cmdi/ANY "" []
              (comp
               (fn [{:keys [params globals puppetdb-query]}]
@@ -280,14 +324,12 @@
 (pls/defn-validated node-routes :- bidi-schema/RoutePair
   [version :- s/Keyword]
   (extract-query
+   typical-params
    (cmdi/routes
     (cmdi/ANY "" []
               (create-query-handler version "nodes" http-q/restrict-query-to-active-nodes))
     (cmdi/context ["/" (route-param :node)]
-                  (cmdi/ANY "" []
-                            (-> (node-status version)
-                                (validate-query-params {:optional ["pretty"]})))
-
+                  (cmdi/ANY "" [] (node-status version))
                   (cmdi/context "/facts"
                                 (-> (facts-routes version)
                                     (append-handler http-q/restrict-query-to-node)
@@ -301,37 +343,42 @@
   [version :- s/Keyword]
   (cmdi/routes
    (extract-query
+    typical-params
     (cmdi/ANY "" []
               (create-query-handler version "environments")))
    (cmdi/context ["/" (route-param :environment)]
                  (cmdi/ANY "" []
                    (validate-query-params (environment-status version)
-                                          {:optional ["pretty"]}))
+                                          (merge-param-specs global-params
+                                                             pretty-params)))
 
                  (wrap-with-parent-check
                   (cmdi/routes
                    (extract-query
+                    typical-params
                     (cmdi/context "/facts"
                                   (-> (facts-routes version)
                                       (append-handler http-q/restrict-query-to-environment))))
 
                    (extract-query
+                    typical-params
                     (cmdi/context "/resources"
                                   (-> (resources-routes version)
                                       (append-handler http-q/restrict-query-to-environment))))
 
                    (extract-query
+                    typical-params
                     (cmdi/context "/reports"
                                   (-> (reports-routes version)
                                       (append-handler http-q/restrict-query-to-environment))))
 
                    (extract-query
-                    {:optional (concat
-                                ["query"
+                    (merge-param-specs
+                     typical-params
+                     {:optional ["query"
                                  "distinct_resources"
                                  "distinct_start_time"
-                                 "distinct_end_time"]
-                                paging/query-params)}
+                                 "distinct_end_time"]})
 
                     (cmdi/context "/events"
                                   (-> (events-routes version)
@@ -342,25 +389,30 @@
   [version :- s/Keyword]
   (cmdi/routes
    (extract-query
+    typical-params
     (cmdi/ANY "" []
               (create-query-handler version "producers")))
    (cmdi/context ["/" (route-param :producer)]
                  (cmdi/ANY "" []
                    (validate-query-params (producer-status version)
-                                          {:optional ["pretty"]}))
+                                          (merge-param-specs global-params
+                                                             pretty-params)))
                  (wrap-with-parent-check
                   (cmdi/routes
                    (extract-query
+                    typical-params
                     (cmdi/context "/factsets"
                                   (-> (factset-routes version)
                                       (append-handler http-q/restrict-query-to-producer))))
 
                   (extract-query
+                    typical-params
                     (cmdi/context "/catalogs"
                                   (-> (catalog-routes version)
                                       (append-handler http-q/restrict-query-to-producer))))
 
                    (extract-query
+                    typical-params
                     (cmdi/context "/reports"
                                   (-> (reports-routes version)
                                       (append-handler http-q/restrict-query-to-producer)))))
@@ -369,33 +421,36 @@
 (pls/defn-validated fact-contents-routes :- bidi-schema/RoutePair
   [version :- s/Keyword]
   (extract-query
+   typical-params
    (cmdi/ANY "" []
              (create-query-handler version "fact_contents" http-q/restrict-query-to-active-nodes))))
 
 (pls/defn-validated fact-path-routes :- bidi-schema/RoutePair
   [version :- s/Keyword]
   (extract-query
+   typical-params
    (cmdi/ANY "" []
              (create-query-handler version "fact_paths"))))
 
 (pls/defn-validated event-counts-routes :- bidi-schema/RoutePair
   [version :- s/Keyword]
   (extract-query
-   {:required ["summarize_by"]
-    :optional (concat ["counts_filter" "count_by"
-                       "distinct_resources" "distinct_start_time"
-                       "distinct_end_time"]
-                      paging/query-params)}
+   (merge-param-specs typical-params
+                      {:required ["summarize_by"]
+                       :optional ["counts_filter" "count_by"
+                                  "distinct_resources" "distinct_start_time"
+                                  "distinct_end_time"]})
    (cmdi/ANY "" []
              (create-query-handler version "event_counts"))))
 
 (pls/defn-validated agg-event-counts-routes :- bidi-schema/RoutePair
   [version :- s/Keyword]
   (extract-query
-   {:required ["summarize_by"]
-    :optional ["query" "counts_filter" "count_by"
-               "distinct_resources" "distinct_start_time"
-               "distinct_end_time"]}
+   (merge-param-specs global-params
+                      {:required ["summarize_by"]
+                       :optional ["query" "counts_filter" "count_by"
+                                  "distinct_resources" "distinct_start_time"
+                                  "distinct_end_time"]})
    (cmdi/ANY "" []
              (create-query-handler version
                                    "aggregate_event_counts"))))
