@@ -34,6 +34,7 @@
    (s/optional-key :ast_only) (s/maybe s/Bool)
    (s/optional-key :include_total) (s/maybe s/Bool)
    (s/optional-key :pretty) (s/maybe s/Bool)
+   (s/optional-key :include_package_inventory) (s/maybe s/Bool)
    (s/optional-key :order_by) (s/maybe [[(s/one s/Keyword "field")
                                          (s/one (s/enum :ascending :descending) "order")]])
    (s/optional-key :distinct_resources) (s/maybe s/Bool)
@@ -267,6 +268,7 @@
       (update-when [:offset] parse-offset)
       (update-when [:include_total] coerce-to-boolean)
       (update-when [:pretty] coerce-to-boolean)
+      (update-when [:include_package_inventory] coerce-to-boolean)
       (update-when [:distinct_resources] coerce-to-boolean)))
 
 (defn get-req->query
@@ -278,6 +280,7 @@
       (update-when ["order_by"] parse-order-by-json)
       (update-when ["counts_filter"] json/parse-strict-string true)
       (update-when ["pretty"] coerce-to-boolean)
+      (update-when ["include_package_inventory"] coerce-to-boolean)
       keywordize-keys))
 
 (defn post-req->query
@@ -312,8 +315,7 @@
      (handler
       (if puppetdb-query
         req
-        (let [param-spec (update param-spec :optional conj "pretty")
-              query-map (create-query-map req param-spec parse-fn)
+        (let [query-map (create-query-map req param-spec parse-fn)
               pretty-print (:pretty query-map
                                     (get-in req [:globals :pretty-print]))]
           (-> req
@@ -362,11 +364,12 @@
 (defn narrow-globals
   "Reduces the number of globals to limit their reach in the codebase"
   [globals]
-  (select-keys globals [:scf-read-db :warn-experimental :url-prefix :pretty-print]))
+  (select-keys globals [:scf-read-db :warn-experimental :url-prefix :pretty-print :node-purge-ttl]))
 
 (defn valid-query?
-  [scf-read-db version query-map]
-  (let [{:keys [remaining-query entity query-options]} (qeng/user-query->engine-query version query-map)]
+  [scf-read-db version query-map query-options]
+  (let [{:keys [remaining-query entity query-options]}
+        (qeng/user-query->engine-query version query-map (:node-purge-ttl query-options))]
     (jdbc/with-db-connection scf-read-db
       (when (qeng/query->sql remaining-query entity version query-options)
         true))))
@@ -374,8 +377,9 @@
 (defn query-handler
   [version]
   (fn [{:keys [params globals puppetdb-query]}]
-    (if (and (:ast_only puppetdb-query) (valid-query? (:scf-read-db globals) version puppetdb-query))
-      (http/json-response (:query puppetdb-query))
-      (qeng/produce-streaming-body version
-                              (validate-distinct-options! (merge (keywordize-keys params) puppetdb-query))
-                              (narrow-globals globals)))))
+    (let [query-options (narrow-globals globals)]
+      (if (and (:ast_only puppetdb-query) (valid-query? (:scf-read-db globals) version puppetdb-query query-options))
+        (http/json-response (:query puppetdb-query))
+        (qeng/produce-streaming-body version
+                                (validate-distinct-options! (merge (keywordize-keys params) puppetdb-query))
+                                query-options)))))
