@@ -11,13 +11,14 @@
             [puppetlabs.puppetdb.meta :as meta]
             [puppetlabs.puppetdb.admin :as admin]
             [puppetlabs.puppetdb.http.command :as cmd]
-            [clj-http.client :as client]
             [puppetlabs.puppetdb.http.server :as server]
             [clojure.tools.logging :as log]
             [puppetlabs.puppetdb.config :as conf]
             [puppetlabs.puppetdb.middleware :as mid]
             [puppetlabs.kitchensink.core :as ks]
-            [puppetlabs.puppetdb.status :as pdb-status]))
+            [puppetlabs.puppetdb.status :as pdb-status]
+            [puppetlabs.i18n.core :refer [trs tru]]
+            [puppetlabs.puppetdb.dashboard :as dashboard]))
 
 
 (defn resource-request-handler [req]
@@ -26,15 +27,15 @@
 (defn maint-mode-handler [maint-mode-fn]
   (fn [req]
     (when (maint-mode-fn)
-      (log/info "HTTP request received while in maintenance mode")
+      (log/info (trs "HTTP request received while in maintenance mode"))
       {:status 503
-       :body "PuppetDB is currently down. Try again later."})))
+       :body (tru "PuppetDB is currently down. Try again later.")})))
 
 (defn wrap-with-context [uri route]
   (compojure/context uri [] route))
 
 (defn pdb-core-routes [defaulted-config get-shared-globals enqueue-command-fn
-                       query-fn clean-fn]
+                       query-fn clean-fn delete-node-fn]
   (let [db-cfg #(select-keys (get-shared-globals) [:scf-read-db])]
     (map #(apply wrap-with-context %)
          (partition
@@ -45,6 +46,7 @@
                       rreq/request-url
                       (format "%s/dashboard/index.html")
                       rr/redirect))
+           "/dashboard" (dashboard/build-app dashboard/default-meter-defs)
            "/meta" (meta/build-app db-cfg defaulted-config)
            "/cmd" (cmd/command-app get-shared-globals
                                    enqueue-command-fn
@@ -54,7 +56,8 @@
            "/admin" (admin/build-app enqueue-command-fn
                                      query-fn
                                      db-cfg
-                                     clean-fn)]))))
+                                     clean-fn
+                                     delete-node-fn)]))))
 
 (defn pdb-app [root maint-mode-fn app-routes]
   (compojure/context root []
@@ -62,7 +65,7 @@
     (maint-mode-handler maint-mode-fn)
     (apply compojure/routes
            (concat app-routes
-                   [(route/not-found "Not Found")]))))
+                   [(route/not-found (tru "Not Found"))]))))
 
 (defprotocol MaintenanceMode
   (enable-maint-mode [this])
@@ -92,7 +95,7 @@
 
 (tk/defservice pdb-routing-service
   [[:WebroutingService add-ring-handler get-route]
-   [:PuppetDBServer clean shared-globals query set-url-prefix]
+   [:PuppetDBServer clean delete-node shared-globals query set-url-prefix]
    [:PuppetDBCommandDispatcher enqueue-command]
    [:MaintenanceMode enable-maint-mode maint-mode? disable-maint-mode]
    [:DefaultedConfig get-config]
@@ -107,7 +110,7 @@
               cert-whitelist (get-in config [:puppetdb :certificate-whitelist])]
           (set-url-prefix query-prefix)
 
-          (log/info "Starting PuppetDB, entering maintenance mode")
+          (log/info (trs "Starting PuppetDB, entering maintenance mode"))
           (add-ring-handler
            this
            (-> (pdb-app context-root
@@ -116,7 +119,8 @@
                                          augmented-globals
                                          enqueue-command
                                          query
-                                         clean))
+                                         clean
+                                         delete-node))
                (mid/wrap-cert-authn cert-whitelist)
                mid/wrap-with-puppetdb-middleware))
 
@@ -127,8 +131,9 @@
                                              (pdb-status/status-details config shared-globals maint-mode?)))))
         context)
   (start [this context]
-         (log/info "PuppetDB finished starting, disabling maintenance mode")
-         (disable-maint-mode)
+         (when-not (get-in (get-config) [:global :upgrade-and-exit?])
+           (log/info (trs "PuppetDB finished starting, disabling maintenance mode"))
+           (disable-maint-mode))
          context)
 
   (stop [this context]
