@@ -11,78 +11,108 @@
   clojure.tools.reader
   (:refer-clojure :exclude [read read-line read-string char
                             default-data-readers *default-data-reader-fn*
-                            *read-eval* *data-readers*])
-  (:use clojure.tools.reader.reader-types
-        [clojure.tools.reader.impl utils commons])
-  (:require [clojure.tools.reader.default-data-readers :as data-readers])
+                            *read-eval* *data-readers* *suppress-read*])
+  (:require [clojure.tools.reader.reader-types :refer
+             [read-char reader-error unread peek-char indexing-reader?
+              get-line-number get-column-number get-file-name string-push-back-reader
+              log-source]]
+            [clojure.tools.reader.impl.utils :refer :all] ;; [char ex-info? whitespace? numeric? desugar-meta]
+            [clojure.tools.reader.impl.commons :refer :all]
+            [clojure.tools.reader.default-data-readers :as data-readers])
   (:import (clojure.lang PersistentHashSet IMeta
                          RT Symbol Reflector Var IObj
                          PersistentVector IRecord Namespace)
-           java.lang.reflect.Constructor))
+           java.lang.reflect.Constructor
+           (java.util regex.Pattern List LinkedList)))
 
 ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
 ;; helpers
 ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
 
-(declare read macros dispatch-macros
+(declare ^:private read*
+         macros dispatch-macros
          ^:dynamic *read-eval*
          ^:dynamic *data-readers*
          ^:dynamic *default-data-reader-fn*
+         ^:dynamic *suppress-read*
          default-data-readers)
 
+(defn ^:private ns-name* [x]
+  (if (instance? Namespace x)
+    (name (ns-name x))
+    (name x)))
+
 (defn- macro-terminating? [ch]
   (case ch
     (\" \; \@ \^ \` \~ \( \) \[ \] \{ \} \\) true
     false))
 
 (defn- ^String read-token
+  "Read in a single logical token from the reader"
   [rdr initch]
   (if-not initch
     (reader-error rdr "EOF while reading")
-    (loop [sb (doto (StringBuilder.) (.append initch))
-           ch (read-char rdr)]
+    (loop [sb (StringBuilder.) ch initch]
       (if (or (whitespace? ch)
               (macro-terminating? ch)
               (nil? ch))
-        (do (unread rdr ch)
+        (do (when ch
+              (unread rdr ch))
             (str sb))
         (recur (.append sb ch) (read-char rdr))))))
 
 (declare read-tagged)
 
 (defn- read-dispatch
-  [rdr _]
+  [rdr _ opts pending-forms]
   (if-let [ch (read-char rdr)]
     (if-let [dm (dispatch-macros ch)]
-      (dm rdr ch)
-      (if-let [obj (read-tagged (doto rdr (unread ch)) ch)] ;; ctor reader is implemented as a taggged literal
-        obj
-        (reader-error rdr "No dispatch macro for " ch)))
+      (dm rdr ch opts pending-forms)
+      (read-tagged (doto rdr (unread ch)) ch opts pending-forms)) ;; ctor reader is implemented as a tagged literal
     (reader-error rdr "EOF while reading character")))
 
 (defn- read-unmatched-delimiter
-  [rdr ch]
+  [rdr ch opts pending-forms]
   (reader-error rdr "Unmatched delimiter " ch))
 
 ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
 ;; readers
 ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
 
+(defn read-regex
+  [rdr ch opts pending-forms]
+  (let [sb (StringBuilder.)]
+    (loop [ch (read-char rdr)]
+      (if (identical? \" ch)
+        (Pattern/compile (str sb))
+        (if (nil? ch)
+          (reader-error rdr "EOF while reading regex")
+          (do
+            (.append sb ch )
+            (when (identical? \\ ch)
+              (let [ch (read-char rdr)]
+                (if (nil? ch)
+                  (reader-error rdr "EOF while reading regex"))
+                (.append sb ch)))
+            (recur (read-char rdr))))))))
+
 (defn- read-unicode-char
-  ([^String token offset length base]
-     (let [l (+ offset length)]
-       (when-not (== (count token) l)
-         (throw (IllegalArgumentException. (str "Invalid unicode character: \\" token))))
-       (loop [i offset uc 0]
-         (if (== i l)
-           (char uc)
-           (let [d (Character/digit (int (nth token i)) (int base))]
-             (if (== d -1)
-               (throw (IllegalArgumentException. (str "Invalid digit: " (nth token i))))
-               (recur (inc i) (long (+ d (* uc base))))))))))
+  ([^String token ^long offset ^long length ^long base]
+   (let [l (+ offset length)]
+     (when-not (== (count token) l)
+       (throw (IllegalArgumentException. (str "Invalid unicode character: \\" token))))
+     (loop [i offset uc 0]
+       (if (== i l)
+         (char uc)
+         (let [d (Character/digit (int (nth token i)) (int base))]
+           (if (== d -1)
+             (throw (IllegalArgumentException. (str "Invalid digit: " (nth token i))))
+             (recur (inc i) (long (+ d (* uc base))))))))))
 
   ([rdr initch base length exact?]
-     (loop [i 1 uc (Character/digit (int initch) (int base))]
+   (let [base (long base)
+         length (long length)]
+     (loop [i 1 uc (long (Character/digit (int initch) (int base)))]
        (if (== uc -1)
          (throw (IllegalArgumentException. (str "Invalid digit: " initch)))
          (if-not (== i length)
@@ -99,16 +129,20 @@
                  (if (== d -1)
                    (throw (IllegalArgumentException. (str "Invalid digit: " ch)))
                    (recur (inc i) (long (+ d (* uc base))))))))
-           (char uc))))))
+           (char uc)))))))
 
 (def ^:private ^:const upper-limit (int \uD7ff))
 (def ^:private ^:const lower-limit (int \uE000))
 
 (defn- read-char*
-  [rdr backslash]
+  "Read in a character literal"
+  [rdr backslash opts pending-forms]
   (let [ch (read-char rdr)]
     (if-not (nil? ch)
-      (let [token (read-token rdr ch)
+      (let [token (if (or (macro-terminating? ch)
+                          (whitespace? ch))
+                    (str ch)
+                    (read-token rdr ch))
             token-len (count token)]
         (cond
 
@@ -129,9 +163,6 @@
              (reader-error rdr "Invalid character constant: \\u" (Integer/toString ic 16))
              c))
 
-         (.startsWith token "x")
-         (read-unicode-char token 1 2 16)
-
          (.startsWith token "o")
          (let [len (dec token-len)]
            (if (> len 3)
@@ -144,32 +175,40 @@
          :else (reader-error rdr "Unsupported character: \\" token)))
       (reader-error rdr "EOF while reading character"))))
 
+(defn ^:private starting-line-col-info [rdr]
+  (when (indexing-reader? rdr)
+    [(get-line-number rdr) (int (dec (int (get-column-number rdr))))]))
+
+(defn ^:private ending-line-col-info [rdr]
+  (when (indexing-reader? rdr)
+    [(get-line-number rdr) (get-column-number rdr)]))
+
+(defonce ^:private READ_EOF (Object.))
+(defonce ^:private READ_FINISHED (Object.))
+
+(def ^:dynamic *read-delim* false)
 (defn- ^PersistentVector read-delimited
-  [delim rdr recursive?]
-  (let [first-line (when (indexing-reader? rdr)
-                     (get-line-number rdr))
+  "Reads and returns a collection ended with delim"
+  [delim rdr opts pending-forms]
+  (let [[start-line start-column] (starting-line-col-info rdr)
         delim (char delim)]
-    (loop [a (transient [])]
-      (if-let [ch (read-past whitespace? rdr)]
-        (if (identical? delim (char ch))
-          (persistent! a)
-          (if-let [macrofn (macros ch)]
-            (let [mret (log-source-unread rdr
-                         (macrofn rdr ch))]
-              (recur (if-not (identical? mret rdr) (conj! a mret) a)))
-            (let [o (read (doto rdr (unread ch)) true nil recursive?)]
-              (recur (if-not (identical? o rdr) (conj! a o) a)))))
-        (reader-error rdr "EOF while reading"
-                      (when first-line
-                        (str ", starting at line " first-line)))))))
+    (binding [*read-delim* true]
+      (loop [a (transient [])]
+        (let [form (read* rdr false READ_EOF delim opts pending-forms)]
+          (if (identical? form READ_FINISHED)
+            (persistent! a)
+            (if (identical? form READ_EOF)
+              (reader-error rdr "EOF while reading"
+                            (when start-line
+                              (str ", starting at line " start-line " and column " start-column)))
+              (recur (conj! a form)))))))))
 
 (defn- read-list
-  [rdr _]
-  (let [[start-line start-column] (when (indexing-reader? rdr)
-                                    [(get-line-number rdr) (int (dec (get-column-number rdr)))])
-        the-list (read-delimited \) rdr true)
-        [end-line end-column] (when (indexing-reader? rdr)
-                                [(get-line-number rdr) (int (get-column-number rdr))])]
+  "Read in a list, including its location if the reader is an indexing reader"
+  [rdr _ opts pending-forms]
+  (let [[start-line start-column] (starting-line-col-info rdr)
+        the-list (read-delimited \) rdr opts pending-forms)
+        [end-line end-column] (ending-line-col-info rdr)]
     (with-meta (if (empty? the-list)
                  '()
                  (clojure.lang.PersistentList/create the-list))
@@ -183,12 +222,11 @@
           :end-column end-column})))))
 
 (defn- read-vector
-  [rdr _]
-  (let [[start-line start-column] (when (indexing-reader? rdr)
-                                    [(get-line-number rdr) (int (dec (get-column-number rdr)))])
-        the-vector (read-delimited \] rdr true)
-        [end-line end-column] (when (indexing-reader? rdr)
-                                [(get-line-number rdr) (int (get-column-number rdr))])]
+  "Read in a vector, including its location if the reader is an indexing reader"
+  [rdr _ opts pending-forms]
+  (let [[start-line start-column] (starting-line-col-info rdr)
+        the-vector (read-delimited \] rdr opts pending-forms)
+        [end-line end-column] (ending-line-col-info rdr)]
     (with-meta the-vector
       (when start-line
         (merge
@@ -200,13 +238,12 @@
           :end-column end-column})))))
 
 (defn- read-map
-  [rdr _]
-  (let [[start-line start-column] (when (indexing-reader? rdr)
-                                    [(get-line-number rdr) (int (dec (get-column-number rdr)))])
-        the-map (read-delimited \} rdr true)
+  "Read in a map, including its location if the reader is an indexing reader"
+  [rdr _ opts pending-forms]
+  (let [[start-line start-column] (starting-line-col-info rdr)
+        the-map (read-delimited \} rdr opts pending-forms)
         map-count (count the-map)
-        [end-line end-column] (when (indexing-reader? rdr)
-                                [(get-line-number rdr) (int (get-column-number rdr))])]
+        [end-line end-column] (ending-line-col-info rdr)]
     (when (odd? map-count)
       (reader-error rdr "Map literal must contain an even number of forms"))
     (with-meta
@@ -223,15 +260,15 @@
           :end-column end-column})))))
 
 (defn- read-number
-  [reader initch]
+  [rdr initch]
   (loop [sb (doto (StringBuilder.) (.append initch))
-         ch (read-char reader)]
+         ch (read-char rdr)]
     (if (or (whitespace? ch) (macros ch) (nil? ch))
       (let [s (str sb)]
-        (unread reader ch)
+        (unread rdr ch)
         (or (match-number s)
-            (reader-error reader "Invalid number format [" s "]")))
-      (recur (doto sb (.append ch)) (read-char reader)))))
+            (reader-error rdr "Invalid number format [" s "]")))
+      (recur (doto sb (.append ch)) (read-char rdr)))))
 
 (defn- escape-char [sb rdr]
   (let [ch (read-char rdr)]
@@ -247,10 +284,6 @@
            (if (== -1 (Character/digit (int ch) 16))
              (reader-error rdr "Invalid unicode escape: \\u" ch)
              (read-unicode-char rdr ch 16 4 true)))
-      \x (let [ch (read-char rdr)]
-           (if (== -1 (Character/digit (int ch) 16))
-             (reader-error rdr "Invalid unicode escape: \\x" ch)
-             (read-unicode-char rdr ch 16 2 true)))
       (if (numeric? ch)
         (let [ch (read-unicode-char rdr ch 8 3 false)]
           (if (> (int ch) 0337)
@@ -259,7 +292,7 @@
         (reader-error rdr "Unsupported escape character: \\" ch)))))
 
 (defn- read-string*
-  [reader _]
+  [reader _ opts pending-forms]
   (loop [sb (StringBuilder.)
          ch (read-char reader)]
     (case ch
@@ -271,8 +304,7 @@
 
 (defn- read-symbol
   [rdr initch]
-  (let [[line column] (when (indexing-reader? rdr)
-                        [(get-line-number rdr) (int (dec (get-column-number rdr)))])]
+  (let [[line column] (starting-line-col-info rdr)]
     (when-let [token (read-token rdr initch)]
       (case token
 
@@ -291,10 +323,11 @@
                   (merge
                    (when-let [file (get-file-name rdr)]
                      {:file file})
-                   {:line line
-                    :column column
-                    :end-line (get-line-number rdr)
-                    :end-column (int (inc (get-column-number rdr)))}))))
+                   (let [[end-line end-column] (ending-line-col-info rdr)]
+                     {:line line
+                      :column column
+                      :end-line end-line
+                      :end-column end-column})))))
             (reader-error rdr "Invalid token: " token))))))
 
 (def ^:dynamic *alias-map*
@@ -310,7 +343,7 @@
       (find-ns sym)))
 
 (defn- read-keyword
-  [reader initch]
+  [reader initch opts pending-forms]
   (let [ch (read-char reader)]
     (if-not (whitespace? ch)
       (let [token (read-token reader ch)
@@ -330,19 +363,20 @@
       (reader-error reader "Invalid token: :"))))
 
 (defn- wrapping-reader
+  "Returns a function which wraps a reader in a call to sym"
   [sym]
-  (fn [rdr _]
-    (list sym (read rdr true nil true))))
+  (fn [rdr _ opts pending-forms]
+    (list sym (read* rdr true nil opts pending-forms))))
 
 (defn- read-meta
-  [rdr _]
+  "Read metadata and return the following object with the metadata applied"
+  [rdr _ opts pending-forms]
   (log-source rdr
-    (let [[line column] (when (indexing-reader? rdr)
-                          [(get-line-number rdr) (int (dec (get-column-number rdr)))])
-          m (desugar-meta (read rdr true nil true))]
+    (let [[line column] (starting-line-col-info rdr)
+          m (desugar-meta (read* rdr true nil opts pending-forms))]
       (when-not (map? m)
         (reader-error rdr "Metadata must be Symbol, Keyword, String or Map"))
-      (let [o (read rdr true nil true)]
+      (let [o (read* rdr true nil opts pending-forms)]
         (if (instance? IMeta o)
           (let [m (if (and line (seq? o))
                     (assoc m :line line :column column)
@@ -353,29 +387,151 @@
           (reader-error rdr "Metadata can only be applied to IMetas"))))))
 
 (defn- read-set
-  [rdr _]
-  (PersistentHashSet/createWithCheck (read-delimited \} rdr true)))
+  [rdr _ opts pending-forms]
+  (let [[start-line start-column] (starting-line-col-info rdr)
+        ;; subtract 1 from start-column so it includes the # in the leading #{
+        start-column (if start-column (int (dec (int start-column))))
+        the-set (PersistentHashSet/createWithCheck (read-delimited \} rdr opts pending-forms))
+        [end-line end-column] (ending-line-col-info rdr)]
+    (with-meta the-set
+      (when start-line
+        (merge
+         (when-let [file (get-file-name rdr)]
+           {:file file})
+         {:line start-line
+          :column start-column
+          :end-line end-line
+          :end-column end-column})))))
 
 (defn- read-discard
-  [rdr _]
-  (read rdr true nil true)
-  rdr)
+  "Read and discard the first object from rdr"
+  [rdr _ opts pending-forms]
+  (doto rdr
+    (read* true nil opts pending-forms)))
+
+(def ^:private RESERVED_FEATURES #{:else :none})
+
+(defn- has-feature?
+  [rdr feature opts]
+  (if (keyword? feature)
+    (or (= :default feature) (contains? (get opts :features) feature))
+    (reader-error rdr (str "Feature should be a keyword: " feature))))
+
+(defn- check-eof-error
+  [form rdr ^long first-line]
+  (when (identical? form READ_EOF)
+    (if (< first-line 0)
+      (reader-error rdr "EOF while reading")
+      (reader-error rdr "EOF while reading, starting at line " first-line))))
+
+(defn- check-reserved-features
+  [rdr form]
+  (when (get RESERVED_FEATURES form)
+    (reader-error rdr (str "Feature name " form " is reserved"))))
+
+(defn- check-invalid-read-cond
+  [form rdr ^long first-line]
+  (when (identical? form READ_FINISHED)
+    (if (< first-line 0)
+      (reader-error rdr "read-cond requires an even number of forms")
+      (reader-error rdr (str "read-cond starting on line " first-line " requires an even number of forms")))))
+
+(defn- read-suppress
+  "Read next form and suppress. Return nil or READ_FINISHED."
+  [first-line rdr opts pending-forms]
+  (binding [*suppress-read* true]
+    (let [form (read* rdr false READ_EOF \) opts pending-forms)]
+      (check-eof-error form rdr first-line)
+      (when (identical? form READ_FINISHED)
+        READ_FINISHED))))
+
+(def ^:private NO_MATCH (Object.))
+
+(defn- match-feature
+  "Read next feature. If matched, read next form and return.
+   Otherwise, read and skip next form, returning READ_FINISHED or nil."
+  [first-line rdr opts pending-forms]
+  (let [feature (read* rdr false READ_EOF \) opts pending-forms)]
+    (check-eof-error feature rdr first-line)
+    (if (= feature READ_FINISHED)
+      READ_FINISHED
+      (do
+        (check-reserved-features rdr feature)
+        (if (has-feature? rdr feature opts)
+          ;; feature matched, read selected form
+          (doto (read* rdr false READ_EOF \) opts pending-forms)
+            (check-eof-error rdr first-line)
+            (check-invalid-read-cond rdr first-line))
+          ;; feature not matched, ignore next form
+          (or (read-suppress first-line rdr opts pending-forms)
+              NO_MATCH))))))
+
+(defn- read-cond-delimited
+  [rdr splicing opts pending-forms]
+  (let [first-line (if (indexing-reader? rdr) (get-line-number rdr) -1)
+        result (loop [matched NO_MATCH
+                      finished nil]
+                 (cond
+                  ;; still looking for match, read feature+form
+                  (identical? matched NO_MATCH)
+                  (let [match (match-feature first-line rdr opts pending-forms)]
+                    (if (identical? match READ_FINISHED)
+                      READ_FINISHED
+                      (recur match nil)))
+
+                  ;; found match, just read and ignore the rest
+                  (not (identical? finished READ_FINISHED))
+                  (recur matched (read-suppress first-line rdr opts pending-forms))
+
+                  :else
+                  matched))]
+    (if (identical? result READ_FINISHED)
+      rdr
+      (if splicing
+        (if (instance? List result)
+          (do
+            (.addAll ^List pending-forms 0 ^List result)
+            rdr)
+          (reader-error rdr "Spliced form list in read-cond-splicing must implement java.util.List."))
+        result))))
+
+(defn- read-cond
+  [rdr _ opts pending-forms]
+  (when (not (and opts (#{:allow :preserve} (:read-cond opts))))
+    (throw (RuntimeException. "Conditional read not allowed")))
+  (if-let [ch (read-char rdr)]
+    (let [splicing (= ch \@)
+          ch (if splicing (read-char rdr) ch)]
+      (when splicing
+        (when-not *read-delim*
+          (reader-error rdr "cond-splice not in list")))
+      (if-let [ch (if (whitespace? ch) (read-past whitespace? rdr) ch)]
+        (if (not= ch \()
+          (throw (RuntimeException. "read-cond body must be a list"))
+          (binding [*suppress-read* (or *suppress-read* (= :preserve (:read-cond opts)))]
+            (if *suppress-read*
+              (reader-conditional (read-list rdr ch opts pending-forms) splicing)
+              (read-cond-delimited rdr splicing opts pending-forms))))
+        (reader-error rdr "EOF while reading character")))
+    (reader-error rdr "EOF while reading character")))
 
 (def ^:private ^:dynamic arg-env)
 
-(defn- garg [n]
+(defn- garg
+  "Get a symbol for an anonymous ?argument?"
+  [^long n]
   (symbol (str (if (== -1 n) "rest" (str "p" n))
                "__" (RT/nextID) "#")))
 
 (defn- read-fn
-  [rdr _]
+  [rdr _ opts pending-forms]
   (if (thread-bound? #'arg-env)
     (throw (IllegalStateException. "Nested #()s are not allowed")))
   (binding [arg-env (sorted-map)]
-    (let [form (read (doto rdr (unread \()) true nil true) ;; this sets bindings
+    (let [form (read* (doto rdr (unread \()) true nil opts pending-forms) ;; this sets bindings
           rargs (rseq arg-env)
           args (if rargs
-                 (let [higharg (key (first rargs))]
+                 (let [higharg (long (key ( first rargs)))]
                    (let [args (loop [i 1 args (transient [])]
                                 (if (> i higharg)
                                   (persistent! args)
@@ -388,7 +544,9 @@
                  [])]
       (list 'fn* args form))))
 
-(defn- register-arg [n]
+(defn- register-arg
+  "Registers an argument to the arg-env"
+  [n]
   (if (thread-bound? #'arg-env)
     (if-let [ret (arg-env n)]
       ret
@@ -400,7 +558,7 @@
 (declare read-symbol)
 
 (defn- read-arg
-  [rdr pct]
+  [rdr pct opts pending-forms]
   (if-not (thread-bound? #'arg-env)
     (read-symbol rdr pct)
     (let [ch (peek-char rdr)]
@@ -415,49 +573,26 @@
            (register-arg -1))
 
        :else
-       (let [n (read rdr true nil true)]
+       (let [n (read* rdr true nil opts pending-forms)]
          (if-not (integer? n)
            (throw (IllegalStateException. "Arg literal must be %, %& or %integer"))
            (register-arg n)))))))
 
 (defn- read-eval
-  [rdr _]
+  "Evaluate a reader literal"
+  [rdr _ opts pending-forms]
   (when-not *read-eval*
     (reader-error rdr "#= not allowed when *read-eval* is false"))
-  (let [o (read rdr true nil true)]
-    (if (symbol? o)
-      (RT/classForName (str ^Symbol o))
-      (if (list? o)
-        (let [fs (first o)
-              o (rest o)
-              fs-name (name fs)]
-          (cond
-           (= fs 'var) (let [vs (first o)]
-                         (RT/var (namespace vs) (name vs)))
-           (.endsWith fs-name ".")
-           (let [args (to-array o)]
-             (-> fs-name (subs 0 (dec (count fs-name)))
-                RT/classForName (Reflector/invokeConstructor args)))
-
-           (Compiler/namesStaticMember fs)
-           (let [args (to-array o)]
-             (Reflector/invokeStaticMethod (namespace fs) fs-name args))
-
-           :else
-           (let [v (Compiler/maybeResolveIn *ns* fs)]
-             (if (var? v)
-               (apply v o)
-               (reader-error rdr "Can't resolve " fs)))))
-        (throw (IllegalArgumentException. "Unsupported #= form"))))))
+  (eval (read* rdr true nil opts pending-forms)))
 
 (def ^:private ^:dynamic gensym-env nil)
 
 (defn- read-unquote
-  [rdr comma]
+  [rdr comma opts pending-forms]
   (if-let [ch (peek-char rdr)]
     (if (identical? \@ ch)
-      ((wrapping-reader 'clojure.core/unquote-splicing) (doto rdr read-char) \@)
-      ((wrapping-reader 'clojure.core/unquote) rdr \~))))
+      ((wrapping-reader 'clojure.core/unquote-splicing) (doto rdr read-char) \@ opts pending-forms)
+      ((wrapping-reader 'clojure.core/unquote) rdr \~ opts pending-forms))))
 
 (declare syntax-quote*)
 (defn- unquote-splicing? [form]
@@ -468,7 +603,9 @@
   (and (seq? form)
        (= (first form) 'clojure.core/unquote)))
 
-(defn- expand-list [s]
+(defn- expand-list
+  "Expand a list by resolving its syntax quotes and unquotes"
+  [s]
   (loop [s (seq s) r (transient [])]
     (if s
       (let [item (first s)
@@ -480,7 +617,9 @@
         (recur (next s) ret))
       (seq (persistent! r)))))
 
-(defn- flatten-map [form]
+(defn- flatten-map
+  "Flatten a map into a seq of alternate keys and values"
+  [form]
   (loop [s (seq form) key-vals (transient [])]
     (if s
       (let [e (first s)]
@@ -499,36 +638,48 @@
         (set! gensym-env (assoc gensym-env sym gs))
         gs)))
 
-(defn- resolve-symbol [s]
+(defn ^:dynamic resolve-symbol
+  "Resolve a symbol s into its fully qualified namespace version"
+  [s]
   (if (pos? (.indexOf (name s) "."))
-    s
+    s ;; If there is a period, it is interop
     (if-let [ns-str (namespace s)]
-      (let [^Namespace ns (resolve-ns (symbol ns-str))]
+      (let [ns (resolve-ns (symbol ns-str))]
         (if (or (nil? ns)
-                (= (name (ns-name ns)) ns-str)) ;; not an alias
+                (= (ns-name* ns) ns-str)) ;; not an alias
           s
-          (symbol (name (.name ns)) (name s))))
+          (symbol (ns-name* ns) (name s))))
       (if-let [o ((ns-map *ns*) s)]
         (if (class? o)
           (symbol (.getName ^Class o))
           (if (var? o)
-            (symbol (-> ^Var o .ns .name name) (-> ^Var o .sym name))))
-        (symbol (name (ns-name *ns*)) (name s))))))
+            (symbol (-> ^Var o .ns ns-name*) (-> ^Var o .sym name))))
+        (symbol (ns-name* *ns*) (name s))))))
 
 (defn- add-meta [form ret]
   (if (and (instance? IObj form)
-           (dissoc (meta form) :line :column :end-line :end-column :file))
+           (seq (dissoc (meta form) :line :column :end-line :end-column :file :source)))
     (list 'clojure.core/with-meta ret (syntax-quote* (meta form)))
     ret))
 
 (defn- syntax-quote-coll [type coll]
-  (let [res (list 'clojure.core/seq
-                  (cons 'clojure.core/concat
-                        (expand-list coll)))]
+  ;; We use sequence rather than seq here to fix http://dev.clojure.org/jira/browse/CLJ-1444
+  ;; But because of http://dev.clojure.org/jira/browse/CLJ-1586 we still need to call seq on the form
+  (let [res (list 'clojure.core/sequence
+                  (list 'clojure.core/seq
+                        (cons 'clojure.core/concat
+                              (expand-list coll))))]
     (if type
       (list 'clojure.core/apply type res)
       res)))
 
+(defn map-func
+  "Decide which map type to use, array-map if less than 16 elements"
+  [coll]
+  (if (>= (count coll) 16)
+    'clojure.core/hash-map
+    'clojure.core/array-map))
+
 (defn- syntax-quote* [form]
   (->>
    (cond
@@ -556,34 +707,39 @@
                :else (resolve-symbol form)))))
 
     (unquote? form) (second form)
-    (unquote-splicing? form) (throw (IllegalStateException. "splice not in list"))
+    (unquote-splicing? form) (throw (IllegalStateException. "unquote-splice not in list"))
 
     (coll? form)
     (cond
+
      (instance? IRecord form) form
-     (map? form) (syntax-quote-coll 'clojure.core/hash-map (flatten-map form))
-     (vector? form) (syntax-quote-coll 'clojure.core/vector form)
+     (map? form) (syntax-quote-coll (map-func form) (flatten-map form))
+     (vector? form) (list 'clojure.core/vec (syntax-quote-coll nil form))
      (set? form) (syntax-quote-coll 'clojure.core/hash-set form)
      (or (seq? form) (list? form))
      (let [seq (seq form)]
        (if seq
          (syntax-quote-coll nil seq)
          '(clojure.core/list)))
+
      :else (throw (UnsupportedOperationException. "Unknown Collection type")))
 
     (or (keyword? form)
         (number? form)
         (char? form)
-        (string? form))
+        (string? form)
+        (nil? form)
+        (instance? Boolean form)
+        (instance? Pattern form))
     form
 
     :else (list 'quote form))
    (add-meta form)))
 
 (defn- read-syntax-quote
-  [rdr backquote]
+  [rdr backquote opts pending-forms]
   (binding [gensym-env {}]
-    (-> (read rdr true nil true)
+    (-> (read* rdr true nil opts pending-forms)
       syntax-quote*)))
 
 (defn- macros [ch]
@@ -618,13 +774,10 @@
     \" read-regex
     \! read-comment
     \_ read-discard
+    \? read-cond
     nil))
 
-(defn- read-tagged* [rdr tag f]
-  (let [o (read rdr true nil true)]
-    (f o)))
-
-(defn- read-ctor [rdr class-name]
+(defn- read-ctor [rdr class-name opts pending-forms]
   (when-not *read-eval*
     (reader-error "Record construction syntax can only be used when *read-eval* == true"))
   (let [class (Class/forName (name class-name) false (RT/baseLoader))
@@ -633,7 +786,7 @@
                              \[ [\] :short]
                              \{ [\} :extended]
                              nil)]
-      (let [entries (to-array (read-delimited end-ch rdr true))
+      (let [entries (to-array (read-delimited end-ch rdr opts pending-forms))
             numargs (count entries)
             all-ctors (.getConstructors class)
             ctors-num (count all-ctors)]
@@ -657,18 +810,20 @@
             (Reflector/invokeStaticMethod class "create" (object-array [vals])))))
       (reader-error rdr "Invalid reader constructor form"))))
 
-(defn- read-tagged [rdr initch]
-  (let [tag (read rdr true nil false)]
+(defn- read-tagged [rdr initch opts pending-forms]
+  (let [tag (read* rdr true nil opts pending-forms)]
     (if-not (symbol? tag)
       (reader-error rdr "Reader tag must be a symbol"))
-    (if-let [f (or (*data-readers* tag)
-                   (default-data-readers tag))]
-      (read-tagged* rdr tag f)
-      (if (.contains (name tag) ".")
-        (read-ctor rdr tag)
-        (if-let [f *default-data-reader-fn*]
-          (f tag (read rdr true nil true))
-          (reader-error rdr "No reader function for tag " (name tag)))))))
+    (if *suppress-read*
+      (tagged-literal tag (read* rdr true nil opts pending-forms))
+      (if-let [f (or (*data-readers* tag)
+                     (default-data-readers tag))]
+        (f (read* rdr true nil opts pending-forms))
+        (if (.contains (name tag) ".")
+          (read-ctor rdr tag opts pending-forms)
+          (if-let [f *default-data-reader-fn*]
+            (f tag (read* rdr true nil opts pending-forms))
+            (reader-error rdr "No reader function for tag " (name tag))))))))
 
 ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
 ;; Public API
@@ -706,47 +861,38 @@
    will be thrown for the unknown tag."
   nil)
 
+(def ^:dynamic *suppress-read* false)
+
 (def default-data-readers
   "Default map of data reader functions provided by Clojure.
    May be overridden by binding *data-readers*"
   {'inst #'data-readers/read-instant-date
    'uuid #'data-readers/default-uuid-reader})
 
-(defn read
-  "Reads the first object from an IPushbackReader or a java.io.PushbackReader.
-   Returns the object read. If EOF, throws if eof-error? is true.
-   Otherwise returns sentinel. If no stream is providen, *in* will be used.
-
-   ***WARNING***
-   Note that read can execute code (controlled by *read-eval*),
-   and as such should be used only with trusted sources.
-
-   To read data structures only, use clojure.tools.reader.edn/read
-
-   Note that the function signature of clojure.tools.reader/read and
-   clojure.tools.reader.edn/read is not the same for eof-handling"
-  ([] (read *in*))
-  ([reader] (read reader true nil))
-  ([reader eof-error? sentinel] (read reader eof-error? sentinel false))
-  ([reader eof-error? sentinel recursive?]
+(defn ^:private read*
+  ([reader eof-error? sentinel opts pending-forms]
+     (read* reader eof-error? sentinel nil opts pending-forms))
+  ([reader eof-error? sentinel return-on opts pending-forms]
      (when (= :unknown *read-eval*)
        (reader-error "Reading disallowed - *read-eval* bound to :unknown"))
      (try
        (loop []
          (log-source reader
-           (let [ch (read-char reader)]
-             (cond
-              (whitespace? ch) (recur)
-              (nil? ch) (if eof-error? (reader-error reader "EOF") sentinel)
-              (number-literal? reader ch) (read-number reader ch)
-              (comment-prefix? ch) (do (read-comment reader) (recur))
-              :else (let [f (macros ch)]
-                      (if f
-                        (let [res (f reader ch)]
-                          (if (identical? res reader)
-                            (recur)
-                            res))
-                        (read-symbol reader ch)))))))
+           (if (seq pending-forms)
+             (.remove ^List pending-forms 0)
+             (let [ch (read-char reader)]
+               (cond
+                (whitespace? ch) (recur)
+                (nil? ch) (if eof-error? (reader-error reader "EOF") sentinel)
+                (= ch return-on) READ_FINISHED
+                (number-literal? reader ch) (read-number reader ch)
+                :else (let [f (macros ch)]
+                        (if f
+                          (let [res (f reader ch opts pending-forms)]
+                            (if (identical? res reader)
+                              (recur)
+                              res))
+                          (read-symbol reader ch))))))))
        (catch Exception e
          (if (ex-info? e)
            (let [d (ex-data e)]
@@ -768,6 +914,32 @@
                                      :file   (get-file-name reader)}))
                            e)))))))
 
+(defn read
+  "Reads the first object from an IPushbackReader or a java.io.PushbackReader.
+   Returns the object read. If EOF, throws if eof-error? is true.
+   Otherwise returns sentinel. If no stream is providen, *in* will be used.
+
+   Opts is a persistent map with valid keys:
+    :read-cond - :allow to process reader conditionals, or
+                 :preserve to keep all branches
+    :features - persistent set of feature keywords for reader conditionals
+    :eof - on eof, return value unless :eofthrow, then throw.
+           if not specified, will throw
+
+   ***WARNING***
+   Note that read can execute code (controlled by *read-eval*),
+   and as such should be used only with trusted sources.
+
+   To read data structures only, use clojure.tools.reader.edn/read
+
+   Note that the function signature of clojure.tools.reader/read and
+   clojure.tools.reader.edn/read is not the same for eof-handling"
+  {:arglists '([] [reader] [opts reader] [reader eof-error? eof-value])}
+  ([] (read *in* true nil))
+  ([reader] (read reader true nil))
+  ([{eof :eof :as opts :or {eof :eofthrow}} reader] (read* reader (= eof :eofthrow) eof nil opts (LinkedList.)))
+  ([reader eof-error? sentinel] (read* reader eof-error? sentinel nil {} (LinkedList.))))
+
 (defn read-string
   "Reads one object from the string s.
    Returns nil when s is nil or empty.
@@ -780,9 +952,11 @@
 
    Note that the function signature of clojure.tools.reader/read-string and
    clojure.tools.reader.edn/read-string is not the same for eof-handling"
-  [s]
-  (when (and s (not (identical? s "")))
-    (read (string-push-back-reader s) true nil false)))
+  ([s]
+     (read-string {} s))
+  ([opts s]
+     (when (and s (not (identical? s "")))
+       (read opts (string-push-back-reader s)))))
 
 (defmacro syntax-quote
   "Macro equivalent to the syntax-quote reader macro (`)."
