@@ -3,7 +3,8 @@
     [clojure.tools.logging :as log]
     [clojure.tools.nrepl.server :as nrepl]
     [puppetlabs.kitchensink.core :refer [to-bool]]
-    [puppetlabs.trapperkeeper.core :refer [defservice]]))
+    [puppetlabs.trapperkeeper.core :refer [defservice]]
+    [puppetlabs.i18n.core :as i18n]))
 
 
 ;; If no port is specified in the config then 7888 is used
@@ -35,14 +36,14 @@
   [get-in-config]
   (let [{:keys [enabled? port bind handler]} (process-config get-in-config)]
     (if enabled?
-      (do (log/info "Starting nREPL service on" bind "port" port)
-        (nrepl/start-server :port port :bind bind :handler handler))
-      (log/info "nREPL service disabled, not starting"))))
+      (do (log/info (i18n/trs "Starting nREPL service on {0} port {1}" bind port))
+          (nrepl/start-server :port port :bind bind :handler handler))
+      (log/info (i18n/trs "nREPL service disabled, not starting")))))
 
 (defn- shutdown-nrepl
   [nrepl-server]
   (when nrepl-server
-    (log/info "Shutting down nREPL service")
+    (log/info (i18n/trs "Shutting down nREPL service"))
     (nrepl/stop-server nrepl-server)))
 
 (defservice nrepl-service
@@ -58,11 +59,11 @@
    If no port is specified then the default port of 7888 is used."
   [[:ConfigService get-in-config]]
   (init [this context]
-        (let [nrepl-server (startup-nrepl get-in-config)]
-          (assoc context :nrepl-server nrepl-server)))
+    (let [nrepl-server (startup-nrepl get-in-config)]
+      (assoc context :nrepl-server nrepl-server)))
   (stop [this context]
-        (shutdown-nrepl (context :nrepl-server))
-        context))
+    (shutdown-nrepl (context :nrepl-server))
+    context))
 
 
 
