@@ -1,12 +1,17 @@
 (ns puppetlabs.trapperkeeper.services.webserver.jetty9-config
-  (:import [java.security KeyStore]
+  (:import (java.security KeyStore)
            (java.io FileInputStream)
+           (java.util HashMap)
+           (ch.qos.logback.access PatternLayout)
+           (ch.qos.logback.core CoreConstants)
            (org.eclipse.jetty.server.handler RequestLogHandler)
            (org.eclipse.jetty.server Server)
            (org.codehaus.janino ScriptEvaluator)
            (org.codehaus.commons.compiler CompileException)
            (java.lang.reflect InvocationTargetException)
-           (com.puppetlabs.trapperkeeper.services.webserver.jetty9.utils LifeCycleImplementingRequestLogImpl))
+           (com.puppetlabs.trapperkeeper.services.webserver.jetty9.utils
+             LifeCycleImplementingRequestLogImpl
+             MDCAccessLogConverter MDCRequestLogHandler))
   (:require [clojure.tools.logging :as log]
             [clojure.string :as str]
             [me.raynes.fs :as fs]
@@ -68,6 +73,7 @@
    "TLS_RSA_WITH_AES_128_CBC_SHA"])
 (def default-protocols ["TLSv1" "TLSv1.1" "TLSv1.2"])
 (def default-client-auth :need)
+(def default-allow-renegotiation false)
 
 ;;;
 ;;; JMX
@@ -117,7 +123,8 @@
    (schema/optional-key :gzip-enable)                schema/Bool
    (schema/optional-key :access-log-config)          schema/Str
    (schema/optional-key :shutdown-timeout-seconds)   schema/Int
-   (schema/optional-key :post-config-script)         schema/Str})
+   (schema/optional-key :post-config-script)         schema/Str
+   (schema/optional-key :allow-renegotiation)        schema/Bool})
 
 (def MultiWebserverRawConfigUnvalidated
   {schema/Keyword  WebserverRawConfig})
@@ -174,7 +181,8 @@
    :client-auth                        WebserverSslClientAuth
    (schema/optional-key :ssl-crl-path) (schema/maybe schema/Str)
    :cipher-suites                      [schema/Str]
-   :protocols                          (schema/maybe [schema/Str])})
+   :protocols                          (schema/maybe [schema/Str])
+   (schema/optional-key :allow-renegotiation)     (schema/maybe schema/Bool)})
 
 (def WebserverSslConnector
   (merge
@@ -388,7 +396,9 @@
             :cipher-suites           (get-cipher-suites-config config)
             :protocols               (get-ssl-protocols-config config)
             :client-auth             (get-client-auth! config)
-            :ssl-crl-path            (get-ssl-crl-path! config)})))
+            :ssl-crl-path            (get-ssl-crl-path! config)
+            :allow-renegotiation     (get config :allow-renegotiation 
+                                         default-allow-renegotiation)})))
 
 (schema/defn ^:always-validate
   maybe-add-http-connector :- {(schema/optional-key :http) WebserverConnector
@@ -437,8 +447,14 @@
 (schema/defn ^:always-validate
   init-log-handler :- RequestLogHandler
   [config :- WebserverRawConfig]
-  (let [handler (RequestLogHandler.)
+  (let [handler (MDCRequestLogHandler.)
+        pattern-rules (HashMap.)
         logger (LifeCycleImplementingRequestLogImpl.)]
+    (doseq [pattern ["X" "mdc"]]
+      (.put pattern-rules
+            pattern
+            (.getName MDCAccessLogConverter)))
+    (.putObject logger CoreConstants/PATTERN_RULE_REGISTRY pattern-rules)
     (.setFileName logger (:access-log-config config))
     (.setQuiet logger true)
     (.setRequestLog handler logger)
